
#include "config.h"
#include "types.h"

#include "p0f-db.h"
#include "p0f-inline.h"

#ifdef ENABLE_POSTGRESQL
static PGconn *pg_conn;
int pdb_pgchecksys(_u8 *a, _u8 *os, _u8 *desc);
#endif

#ifdef ENABLE_MYSQL
static MYSQL *mysql_conn;
int pdb_mysqlchecksys(_u8 *a, _u8 *os, _u8 *desc);
#endif

#ifdef NO_STRNDUP
char* temp_strndup (const char* string, size_t n){
 char* newstring;
 size_t len=strlen(string);
 if ( len>n ) 
    len=n;
 newstring = (char*)malloc(len+1);
 if (newstring==NULL) return NULL;
  newstring[len] = '\0';
  return (char *)memcpy(newstring, string, len);
}
#define strndup temp_strndup
#endif

char *pdb_escape(char *str);
void pdb_parseconfig();

/*-----*/

#ifdef ENABLE_POSTGRESQL

void pdb_pgout(_u8 *a, _u16 sp, _u8 *os, _u8 *desc, struct timeval pts)
{
  char     query[PDB_QUERY_LEN];
  PGresult *res;
  char *escaped;

  if (!pg_conn)
  {
    /* debug("PDB_PG: Not connected to DB.\n"); */
    return;
  }
  
  /* Start a transaction block */
  res = PQexec(pg_conn, "BEGIN");
  if (PQresultStatus(res) != PGRES_COMMAND_OK)
  {
    debug("PDB_PG: BEGIN command failed: %s", PQerrorMessage(pg_conn));
    PQclear(res);
    PQfinish(pg_conn);
    fatal("PDB_PG: Exiting.\n");
  }
  PQclear(res);

  /* do insert/update */

  escaped = pdb_escape(os);
  if (!pdb_pgchecksys(a, os, desc))
  {  /* IP not in DB -- make a new record */
    snprintf(query, PDB_QUERY_LEN, 
      "INSERT INTO system (ip_addr, name, first_tstamp, last_tstamp) VALUES ('%d.%d.%d.%d', '%.32s %.64s', '%.32s', '%.32s');",
        a[0], a[1], a[2], a[3], escaped, desc, asctime(localtime(&pts.tv_sec)), asctime(localtime(&pts.tv_sec)));
  }
  else
  {  /* IP is already in DB -- update time stamp */
    snprintf(query, PDB_QUERY_LEN,
      "UPDATE system SET last_tstamp = '%.32s' WHERE ip_addr = '%d.%d.%d.%d' AND name = '%.32s %.64s'",
        asctime(localtime(&pts.tv_sec)), a[0], a[1], a[2], a[3], escaped, desc);
  }
  free(escaped);
  
  res = PQexec(pg_conn, query);
  if (PQresultStatus(res) != PGRES_COMMAND_OK)
  {
    debug("PDB_PG: Insert/Update failed: %s", PQerrorMessage(pg_conn));
    PQclear(res);
    PQfinish(pg_conn);
    fatal("PDB_PG: Exiting.\n");
  }
  PQclear(res);

  /* end the transaction */
  res = PQexec(pg_conn, "END");
  if (PQresultStatus(res) != PGRES_COMMAND_OK)
  {
    debug("PDB_PG: END command failed: %s", PQerrorMessage(pg_conn));
    PQclear(res);
    PQfinish(pg_conn);
    fatal("PDB_PG: Exiting.\n");
  }
  PQclear(res);
}

/*-----*/

void pdb_pgconnect(const char *conninfo)
{
  pg_conn = PQconnectdb(conninfo);
  if (PQstatus(pg_conn) != CONNECTION_OK)
  {
    debug("PDB_PG: Connection to database '%s' failed.\n", PQdb(pg_conn));
    PQfinish(pg_conn);
    fatal("PDB_PG: Exiting.\n");
  }
  else
    debug("PDB_PG: Connected to '%s' database.\n", PQdb(pg_conn));
}

/*-----*/

void pdb_pgdisconnect()
{
  if (pg_conn)
  {
    debug("PDB_PG: Closing connection to '%s'.\n", PQdb(pg_conn));
    PQfinish(pg_conn);
  }
}

/*-----*/

int pdb_pgchecksys(_u8 *a, _u8 *os, _u8 *desc)
{
  char ipquery[PDB_QUERY_LEN];
  PGresult *res;
  int nTuples;
  char *escaped;
  
  escaped = pdb_escape(os);
  snprintf(ipquery, PDB_QUERY_LEN, 
    "SELECT ip_addr FROM system WHERE ip_addr = '%d.%d.%d.%d' AND name = '%.32s %.64s' LIMIT 1;", 
      a[0], a[1], a[2], a[3], escaped, desc);
  res = PQexec(pg_conn, ipquery);
  if (PQresultStatus(res) != PGRES_TUPLES_OK)
  {
    debug("PDB_PG: Check IP failed: %s", PQerrorMessage(pg_conn));
    PQclear(res);
    PQfinish(pg_conn);
    fatal("PDB_PG: Exiting due to DB failure.\n");
  }
  free(escaped);
  
  /* Return the number of rows selected */
  nTuples = PQntuples(res);
  PQclear(res);

  return nTuples;
}

#endif /* ifdef ENABLE_POSTGRESQL */

/*-----*/

#ifdef ENABLE_MYSQL

void pdb_mysqlout(_u8 *a, _u16 sp, _u8 *os, _u8 *desc, struct timeval pts)
{
  char query[PDB_QUERY_LEN];
  char *escaped;

  if (!mysql_conn)
  {
    /* debug("PDB_MY: Not connected to DB.\n"); */
    return;
  }

  escaped = pdb_escape(os);
  if (!pdb_mysqlchecksys(a, os, desc))
  {  /* IP not in DB -- make a new record */
    snprintf(query, PDB_QUERY_LEN, 
      "INSERT INTO system (ip_addr, name, first_tstamp, last_tstamp) VALUES (INET_ATON('%d.%d.%d.%d'), '%.32s %.64s', FROM_UNIXTIME(%.32ld), FROM_UNIXTIME(%.32ld));",
        a[0], a[1], a[2], a[3], escaped, desc, mktime(localtime(&pts.tv_sec)), mktime(localtime(&pts.tv_sec)));
  }
  else
  {  /* IP is already in DB -- update time stamp */
    snprintf(query, PDB_QUERY_LEN,
      "UPDATE system SET last_tstamp = FROM_UNIXTIME(%.32d) WHERE ip_addr = INET_ATON('%ld.%ld.%ld.%ld') AND name = '%.32s %.64s'",
        mktime(localtime(&pts.tv_sec)), a[0], a[1], a[2], a[3], escaped, desc);
  }
  free(escaped);
  
  if (mysql_query(mysql_conn, query))  /* If this returns non zero -> error occured */
  {
    debug("PDB_MY: Insert/Update failed: %s", mysql_error(mysql_conn));
    mysql_close(mysql_conn);
    fatal("PDB_MY: Exiting.\n");
  }
}

/*-----*/

void pdb_mysqlconnect(const char *conninfo)
{
  #define MCFLEN 128

  char *host = "localhost";
  char *user;
  char *password;
  char *db;
  int port = 3306;

  char *myvar;
  char *myval;
  char *mycon;
  
  void setconnopt(const char *var, const char *val)
  {
    if (strncasecmp(var, "HOST", MCFLEN) == 0)  /* host part */
    {
      host = strndup(val, MCFLEN);
      if (!host)
        fatal("PDB_MY: Out of memory.\n");
    }
    else if (strncasecmp(var, "PORT", MCFLEN) == 0)  /* port # */
    {
      port = atol(val);
    }
    else if (strncasecmp(var, "USER", MCFLEN) == 0)  /* user name */
    {
      user = strndup(val, MCFLEN);
      if (!user)
        fatal("PDB_MY: Out of memory.\n");
    }
    else if (strncasecmp(var, "PASSWORD", MCFLEN) == 0)  /* user password */
    {
      password = strndup(val, MCFLEN);
      if (!password)
        fatal("PDB_MY: Out of memory.\n");
    }
    else if (strncasecmp(var, "DBNAME", MCFLEN) == 0)  /* database name */
    {
      db = strndup(val, MCFLEN);
      if (!db)
        fatal("PDB_MY: Out of memory.\n");
    }
    else  /* unknown keyword in MySQL connection string */
    {
      fatal("PDB_MY: Unknown keyword '%s' in config file.\n", var);
    }
  }
 
  // parse conninfo into separate values
  mycon = strndup(conninfo, PDB_CONF_LINE_LEN);
  if (!mycon)
    fatal("PDB_MY: Out of memory.\n");
  myvar = strtok(mycon, "=");
  myval = strtok(NULL, " ");
  if (myvar && myval) setconnopt(myvar, myval);
  while (myvar && myval)  /* if there're any more arguments, get them */
  {
    myvar = strtok(NULL, "=");
    if (!myvar) break;
    myval = strtok(NULL, " ");
    if (!myval) break;
    setconnopt(myvar, myval);
  }
  // do connect!
  mysql_conn = mysql_init(NULL);
  if (!mysql_conn)
    fatal("PDB_MY: Out of memory.\n");
  mysql_real_connect(mysql_conn, host, user, password, db, port, NULL, 0);
  if (!mysql_conn)  /* connection failed? */
  {
    debug("PDB_MY: Connection to database '%s' failed.\n", mysql_error(mysql_conn));
    mysql_close(mysql_conn);
    fatal("PDB_MY: Exiting.\n");
  }
  else
    debug("PDB_MY: Connected to '%s' database.\n", db);

  
}

/*-----*/

void pdb_mysqldisconnect()
{
  if (mysql_conn)
  {
    debug("PDB_MY: Closing connection to MySQL DB.\n");
    mysql_close(mysql_conn);
  }
}

/*-----*/
int pdb_mysqlchecksys(_u8 *a, _u8 *os, _u8 *desc)
{
  char ipquery[PDB_QUERY_LEN];
  MYSQL_RES *res;
  my_ulonglong rows;
  char *escaped;

  escaped = pdb_escape(os);
  snprintf(ipquery, PDB_QUERY_LEN, 
    "SELECT ip_addr FROM system WHERE ip_addr = INET_ATON('%d.%d.%d.%d') AND name = '%.32s %.64s' LIMIT 1;", 
      a[0], a[1], a[2], a[3], escaped, desc);
  free(escaped);

  if (mysql_query(mysql_conn, ipquery))
  {
    debug("PDB_MY: Check IP failed: %s", mysql_error(mysql_conn));
    mysql_close(mysql_conn);
    fatal("PDB_MY: Exiting due to DB failure.\n");
  }

  res = mysql_store_result(mysql_conn);
  rows = mysql_num_rows(res);
  mysql_free_result(res);

  return rows;
}

#endif /* ifdef ENABLE_MYSQL */

/*-----*/

char *pdb_escape(char *str)
{
/*  Ideas borrowed from SNORT (www.snort.org) 
 *  database output plugin.
 */
  char *buf, *buf_start;

  buf = (char *)malloc(strlen(str) * 2 + 1);
  memset(buf, 0, strlen(str) * 2 + 1);
  buf_start = buf;
  
  while (*str)
  {
    switch (*str)
    {
      case '\'':
        *buf = '\''; buf++;
	*buf = '\'';
      default:
        *buf = *str;
    }
    str++; buf++;
  }
  
  return buf_start;
}

/*-----*/

void pdb_parseconfig()
{
  int  linenum;
  FILE *conff;
  char confline[PDB_CONF_LINE_LEN];
  char *conftok;

  conff = fopen(PDB_CONF_FILE_NAME, "r");
  if (!conff)
  {
    fatal("PDB: Cannot open config file: %s\n", PDB_CONF_FILE_NAME);
  }

  linenum = 0;
  while (fgets(confline, PDB_CONF_LINE_LEN, conff))
  {
    linenum++;
    
    conftok = strtok(confline, " ");  /* get a first token */
    if (conftok && conftok[0] != '#' && conftok[0] != '\n')  /* if this is comment/empty line -- throw it off */
    {
      if (strcasecmp(conftok, "POSTGRESQL") == 0)  /* postgres output? */
      {
        #ifndef ENABLE_POSTGRESQL
        fatal("PDB: %s: PostgreSQL support not compiled.\n", PDB_CONF_FILE_NAME);
	#else
        conftok = strtok(NULL, "\n");
        /* conftok now should be equal to connection info (user, dbname, etc.) */
	pdb_pgconnect(conftok);
        #endif
      }
      else if (strcasecmp(conftok, "MYSQL") == 0)  /* mysql output? */
      {
	#ifndef ENABLE_MYSQL
	fatal("PDB: %s: MySQL support not compiled.\n", PDB_CONF_FILE_NAME);
	#else
	conftok = strtok(NULL, "\n");
	/* conftok now should be equal to connection info (user, dbname, etc.) */
	pdb_mysqlconnect(conftok);
	#endif
      }
      else
      {
        debug("PDB: Uknown keyword found\n");
        fatal("PDB: %s:%d: %s", PDB_CONF_FILE_NAME, linenum, conftok);
      }
    }  /* if (conftok && conftok[0] != '#') */
  }

  fclose(conff);
}

/*-----*/


